import math
from typing import List, Tuple, Union

currentOctave = 0

noteTupleDuration = Tuple[str, int, float, int]
noteTuple = Tuple[str, int, float]
frequencyTuple = Tuple[float, float]
# Note tuple has (Note Letter, Octave, Duration(, repetition times))


# Helper functions:
def isFloat(s: str) -> bool:
    return s.replace(".", "").isdigit()


def appendMultipleTimes(l: list, a: any, i: int) -> list:
    c = 0
    nl = l
    while c < i:
        nl.append(a)
        c += 1
    return nl


def lastCharInString(s: str) -> chr:
    return s[len(s)-1]


def getArrayFromFile(fileName: str) -> list:
    with open(fileName, "r") as file:
        return list(map(lambda s: s.replace("\n", ""), file.readlines()))


# Main functions:
def parseNoteString(note: str) -> noteTupleDuration:
    """Takes a string like "2x 3 E5" and parses it into a note tuple.

    Parameters
    ----------
    note: str
        String representation of note.
            e.g. 2x 3 E5 for 2 repetitions of an E in the 5th octave with
            duration 3

    Returns
    -------
    noteTuple
        NoteTuple with information from noteString
            "2x 3 E5" -> (E, 5, 3, 2)

    """
    global currentOctave
    parts = note.split(" ")
    nums = 1
    duration = 1.0
    currIndex = 0
    if(
        len(parts[currIndex]) > 1 and
        isFloat(parts[currIndex][:-1]) and
        parts[currIndex][-1:] == 'x'
    ):
        nums = int(parts[currIndex][:-1])
        currIndex += 1
    if isFloat(parts[currIndex]) and lastCharInString(parts[currIndex]) != 'x':
        duration = float(parts[currIndex])
        currIndex += 1
    noteString = parts[currIndex]
    if isFloat(noteString[len(noteString) - 1]):
        currentOctave = int(noteString[len(noteString) - 1])
        noteString = noteString[:-1]
    return (noteString, currentOctave, duration, nums)


def parseArrayToNotePair(noteArray: List[str]) -> List[List[noteTuple]]:
    """Takes a list of note strings and returns a list of note tuples.

    Parameters
    ----------
    noteArray : List[str]
        List of note strings.
            e.g.
                [
                    "2x 3 E5",
                    "2 D",
                    "3x C",
                    ...
                ]

    Returns
    -------
    List[noteTuple]
        Parsed list.
            e.g.
                [
                    (E, 5, 3, 2),
                    (D, 5, 2, 1),
                    (C, 5, 1, 3),
                    ...
                ]

    """
    parsedArray = []

    for note in noteArray:
        arr = []
        if "," not in note:
            ns, o, d, n = parseNoteString(note)
            appendMultipleTimes(parsedArray, [(ns, o, d)], n)
            continue
        else:
            splitArr = note.split(",")
            genDuration = parseNoteString(splitArr[0])[2]
            for partialNote in splitArr:
                partialNote = partialNote.strip()
                ns, o, _, n = parseNoteString(partialNote)
                arr.append((ns, o, genDuration))
        parsedArray.append(arr)

    return parsedArray


shiftDict = {
    "A": 0,
    "A#": 1,
    "Bf": 1,
    "B": 2,
    "C": 3,
    "C#": 4,
    "Df": 4,
    "D": 5,
    "D#": 6,
    "Ef": 6,
    "E": 7,
    "F": 8,
    "F#": 9,
    "Gf": 9,
    "G": 10,
    "G#": 11,
    "Af": 11
}

baseFreq = 440  # A4 is 440
expBase = math.pow(2, 1.0/12)


def noteTupleToFrequency(noteT: noteTuple) -> frequencyTuple:
    """Takes a note tuple and returns a frequency tuple for the note.

    Parameters
    ----------
    notePair : noteTuple
        NoteTuple for note.

    Returns
    -------
    frequencyTuple
        Frequency tuple for note.

    """
    note, octave, duration = noteT
    if note == "_":
        return (0, duration)
    semitoneDiff = (octave-4 - 1)*12 + shiftDict[note]
    # -1 to account for the fact that octaves start at 0.
    return (baseFreq * math.pow(expBase, semitoneDiff), duration)


def fileToFrequency(fileName: str) -> List[List[frequencyTuple]]:
    """Short summary.

    Parameters
    ----------
    fileName : str
        Filename of music file.

    Returns
    -------
    List[List[frequencyTuple]]
        Music file parsed into frequencies.

    """
    arr = getArrayFromFile(fileName)
    np = parseArrayToNotePair(arr)
    song = []
    for chord in np:
        freqChord = list(map(noteTupleToFrequency, chord))
        song.append(freqChord)
    return song



from typing import Union

import librosa
import numpy
from scipy.io import wavfile

ndarray = numpy.ndarray
genlist = Union[ndarray, list]


# Helper functions:
def secondsToSamples(seconds: float, sr: int) -> int:
    return int(seconds*sr)


def maxIndex(l: genlist) -> int:
    maxI = 0
    maxVal = l[0]
    for i in range(1, len(l)):
        val = l[i]
        if val > maxVal:
            maxVal = val
            maxI = i
    return maxI


def chunkAverage(l: genlist, chunkSize: int) -> genlist:
    if chunkSize > len(l):
        raise ValueError("ChunkSize must be less than length of list.")
    avgList = []
    currentIndex = 0
    while (currentIndex + chunkSize) < len(l):
        avgList.append(numpy.mean(l[currentIndex:currentIndex + chunkSize]))
        currentIndex += chunkSize
    avgList.append(numpy.mean(l[currentIndex:]))
    return avgList


def loadFromFile(fileName: str) -> (list, int):
    sr, _ = wavfile.read(fileName, "rb")
    wf, _ = librosa.core.load(fileName)
    return (wf, sr)


def getNSecondsFromWaveForm(waveForm: genlist, sr: int, seconds: float) -> list:
    return waveForm[:secondsToSamples(seconds, sr)]


# Main functions:
def condenseToSingleChannel(waveForm: genlist) -> list:
    waveFormList = waveForm.tolist()
    if (isinstance(waveFormList[0], list)):
        waveFormSingleChannel = list(
            map(lambda l: l[0], waveForm)
        )
    else:
        waveFormSingleChannel = waveForm
    return waveFormSingleChannel


def rfftAudible(waveForm: genlist) -> ndarray:
    fft = numpy.fft.rfft(waveForm)
    mfft = []
    for n in range(0, len(fft)):
        if n < 150 or n > 3000:
            mfft.append(0)
        else:
            mfft.append(numpy.sqrt(fft[n].real**2 + fft[n].imag**2))
    return mfft


def startingNote(waveForm: ndarray, sr: int) -> int:
    sc = condenseToSingleChannel(waveForm)
    beatSample = getNSecondsFromWaveForm(sc, sr, 1.5)
    trimmed, _ = librosa.effects.trim(beatSample)
    adRFFT = rfftAudible(trimmed)
    return maxIndex(adRFFT)


def startingNoteFromFile(fileName: str) -> int:
    waveForm, sr = loadFromFile(fileName)
    return startingNote(waveForm, sr)



import os
import sys
from typing import Tuple

import librosa
import numpy as np
from scipy.io import wavfile

from . import tools

analysis = tools.analysis
parseMusic = tools.parseMusic


def getSPB(bpm: float) -> float:
    bps = bpm/60
    return 1.0/bps


def addList(l1: list, l2: list) -> list:
    if len(l1) != len(l2):
        raise ValueError("Lists must be of Equal Length")
    comb = zip(l1, l2)
    return list(map(lambda t: t[0] + t[1], comb))


def mdArrayAverage(md: list) -> list:
    sumArr = md[0]
    for i in range(1, len(md)):
        sumArr = addList(sumArr, md[i])
    return list(map(lambda v: float(v)/len(md), sumArr))


def shiftPitchBy(noteArray: np.ndarray, sr: int, shift: float) -> np.ndarray:
    return librosa.effects.pitch_shift(noteArray, sr, shift)


def speedUpBy(noteArray: np.ndarray, sr: int, factor: float) -> np.ndarray:
    spedUp = librosa.effects.time_stretch(noteArray, factor)
    return spedUp


def speedUpTo(noteArray: np.ndarray, sr: int, shiftTo: float) -> np.ndarray:
    dur = librosa.core.get_duration(noteArray, sr)
    ratio = dur/shiftTo
    spedUp = speedUpBy(noteArray, sr, ratio)
    return spedUp


def pitchTo(noteArray: np.ndarray, sr: int, shiftTo: int, bpm: float) -> np.ndarray:
    spb = getSPB(bpm)

    currentFrequency = analysis.startingNote(noteArray, sr)
    freqRatio = shiftTo/float(currentFrequency)
    shift = np.log2(freqRatio)
    shiftedPitchWF = shiftPitchBy(noteArray, sr, 12*shift)
    return speedUpTo(shiftedPitchWF, sr, spb)

def emptyBeatTrack(likeList: np.ndarray, sr: int, bpm: float):
    spb = getSPB(bpm)
    emptyLike = np.full_like(likeList, .01)
    return speedUpTo(emptyLike, sr, spb)

def sampleIntoSong(sampleTrack: np.ndarray, sr: int, frequencyList: list, bpm: float) -> np.ndarray:
    song = np.array([])
    for freqList in frequencyList:
        nextParts = []
        for freq, duration in freqList:
            if freq != 0:
                nextParts.append(pitchTo(sampleTrack, sr, freq, bpm))
            else:
                nextParts.append(emptyBeatTrack(sampleTrack, sr, bpm))
        avgArr = np.array(mdArrayAverage(nextParts))
        song = np.concatenate((song, speedUpBy(avgArr, sr, 1.0/duration)))
    return song


def sampleWAVFileIntoMusic(
                    wavFileName: str,
                    musicFileName: str,
                    bpm: float
                    ) -> Tuple[np.ndarray, int]:
    sr, _ = wavfile.read(wavFileName)
    waveForm, _ = librosa.core.load(wavFileName, sr)

    waveForm, _ = librosa.effects.trim(waveForm)
    freqList = parseMusic.fileToFrequency(musicFileName)

    wP = sampleIntoSong(waveForm, sr, freqList, bpm=bpm)
    return (wP, sr)


if __name__ == "__main__":
    bpm = 100

    filename1 = sys.argv[1]
    filename2 = sys.argv[2]
    dir1, name1 = os.path.split(filename1)
    dir2, name2 = os.path.split(filename2)
    name1 = name1.replace(".wav", "")
    name2 = name2.replace(".music", "")

    if len(sys.argv) > 3:
        bpm = int(sys.argv[3])

    wP, sr = sampleWAVFileIntoMusic(filename1, filename2, bpm)
    librosa.output.write_wav("{0}_ShiftedTo_{1}.wav".format(name1, name2), wP, sr)
    os.system("play {0}_ShiftedTo_{1}.wav".format(name1, name2))
